(ns tservice.plugins.quartet-dnaseq-report
  (:require [clojure.data.json :as json]
            [clojure.spec.alpha :as s]
            [clojure.tools.logging :as log]
            [spec-tools.core :as st]
            [clojure.java.shell :as shell :refer [sh]]
            [tservice.lib.files :as ff]
            [tservice.lib.fs :as fs-lib]
            [tservice.vendor.multiqc :as mq]
            [tservice.api.config :refer [add-env-to-path]]
            [tservice.api.storage.fs :as fs-api]
            [tservice.api.task :refer [publish-event! make-plugin-metadata make-events-init create-task! update-process!]]))

;;; ------------------------------------------------ Event Specs ------------------------------------------------
(s/def ::name
  (st/spec
   {:spec                string?
    :type                :string
    :description         "The name of the report"
    :swagger/default     ""
    :reason              "Not a valid report name"}))

(s/def ::description
  (st/spec
   {:spec                string?
    :type                :string
    :description         "Description of the report"
    :swagger/default     ""
    :reason              "Not a valid description."}))

(s/def ::filepath
  (st/spec
   {:spec                (s/and string? #(re-matches #"^oss:\/\/[a-zA-Z0-9_\/]+.*" %))
    :type                :string
    :description         "File path for covertor."
    :swagger/default     nil
    :reason              "The filepath must be string."}))

(def quartet-dna-report-params-body
  "A spec for the body parameters."
  (s/keys :req-un [::name ::description ::filepath]
          :opt-un [::description]))

;;; ------------------------------------------------ Event Metadata ------------------------------------------------
(def metadata
  (make-plugin-metadata
   {:name "quartet-dnaseq-report"
    :summary "Generate the QC Report for Quartet DNA-Seq data."
    :params-schema quartet-dna-report-params-body
    :handler (fn [{:keys [name description filepath owner] :as payload}]
               (let [workdir (ff/get-workdir)
                     report-file (fs-lib/join-paths workdir "multireport.html")
                     log-path (fs-lib/join-paths workdir "log")
                     response {:report report-file
                               :log log-path
                               :response-type :data2report}
                     task-id (create-task! {:name name
                                            :description (or description "Generate the QC Report for Quartet DNA-Seq data.")
                                            :payload payload
                                            :owner owner
                                            :plugin-name "quartet-dnaseq-report"
                                            :plugin-type "ReportPlugin"
                                            :plugin-version "v0.1.0"
                                            :response response})]
                 (fs-lib/create-directories! workdir)
                 (spit log-path (json/write-str {:status "Running" :msg ""}))
                 (update-process! task-id 0)
                 (log/info "quartet_dnaseq_report: " filepath name description)
                 (publish-event! "quartet_dnaseq_report"
                                 {:datadir filepath
                                  :parameters {:name name
                                               :description description
                                               :plugin-name "quartet-dnaseq-report"
                                               :plugin-version "v0.1.0"}
                                  :dest-dir workdir
                                  :task-id task-id})
                 response))
    :plugin-type :ReportPlugin
    :response-type :data2report}))

;;; ------------------------------------------------ Event Processing ------------------------------------------------

(defn- decompression-tar
  [filepath]
  (shell/with-sh-env {:PATH   (add-env-to-path "quartet-dnaseq-report")
                      :LC_ALL "en_US.utf-8"
                      :LANG   "en_US.utf-8"}
    (let [command ["bash" "-c"
                   (format "tar -xvf %s -C %s" filepath (fs-lib/parent-path filepath))]
          result  (apply sh command)
          status (if (= (:exit result) 0) "Success" "Error")
          msg (str (:out result) "\n" (:err result))]
      {:status status
       :msg msg})))

(defn- filter-mkdir-copy
  [fmc-datadir fmc-patterns fmc-destdir fmc-newdir]
  (let [files-keep (fs-api/batch-filter-files fmc-datadir fmc-patterns)
        files-keep-dir (fs-lib/join-paths fmc-destdir fmc-newdir)]
    (fs-lib/create-directories! files-keep-dir)
    (fs-api/copy-files! files-keep files-keep-dir {:replace-existing true})))

(defn date
  []
  (.format (java.text.SimpleDateFormat. "yyyy-MM-dd")
           (new java.util.Date)))

(defn- quartet-dnaseq-report!
  "Chaining Pipeline: filter-files -> copy-files -> multiqc."
  [{:keys [datadir parameters dest-dir task-id]}]
  (log/info "Generate quartet dnaseq report: " datadir parameters dest-dir)
  (let [parameters-file (fs-lib/join-paths dest-dir "general_information.json")
        log-path (fs-lib/join-paths dest-dir "log")]
    (try
      (filter-mkdir-copy datadir [".*call-extract_tables/.*.txt"] dest-dir "call-extract_tables")
      (filter-mkdir-copy datadir [".*call-qualimap_D5/.*zip"] dest-dir "call-qualimap_D5")
      (filter-mkdir-copy datadir [".*call-qualimap_D6/.*zip"] dest-dir "call-qualimap_D6")
      (filter-mkdir-copy datadir [".*call-qualimap_F7/.*zip"] dest-dir "call-qualimap_F7")
      (filter-mkdir-copy datadir [".*call-qualimap_M8/.*zip"] dest-dir "call-qualimap_M8")
      (filter-mkdir-copy datadir [".*call-fastqc_D5/.*.(zip|html)"] dest-dir "call-fastqc_D5")
      (filter-mkdir-copy datadir [".*call-fastqc_D6/.*.(zip|html)"] dest-dir "call-fastqc_D6")
      (filter-mkdir-copy datadir [".*call-fastqc_F7/.*.(zip|html)"] dest-dir "call-fastqc_F7")
      (filter-mkdir-copy datadir [".*call-fastqc_M8/.*.(zip|html)"] dest-dir "call-fastqc_M8")
      (filter-mkdir-copy datadir [".*call-merge_mendelian/.*.txt"] dest-dir "call-merge_mendelian")
      (update-process! task-id 10)
      (spit parameters-file (json/write-str parameters))
      (doseq [files-qualimap-tar (fs-api/batch-filter-files dest-dir [".*qualimap.zip"])]
        (decompression-tar files-qualimap-tar))
      (update-process! task-id 50)
      (spit parameters-file (json/write-str {"Report Name" (:name parameters)
                                             "Description" (:description parameters)
                                             "Report Tool" (format "%s-%s"
                                                                   (:plugin-name parameters)
                                                                   (:plugin-version parameters))
                                             "Team" "Quartet Team"
                                             "Date" (date)}))
      (let [result (mq/multiqc dest-dir dest-dir {:template "quartet_dnaseq_report"
                                                  :title "Quartet DNA report"
                                                  :env {:PATH (add-env-to-path "quartet-dnaseq-report")}})
            log (json/write-str result)]
        (log/info "Status: " result)
        (spit log-path log))
      (update-process! task-id 100)
      (catch Exception e
        (update-process! task-id -1)
        (let [log (json/write-str {:status "Error" :msg (.toString e)})]
          (log/info "Status: " log)
          (spit log-path log))))))

;;; --------------------------------------------------- Lifecycle ----------------------------------------------------
(def events-init
  "Automatically called during startup; start event listener for quartet_dnaseq_report events."
  (make-events-init "quartet_dnaseq_report" quartet-dnaseq-report!))
